import re

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Bitly(SiteVisitor):
    NAME = "Bitly"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"bit\.ly/(?P<id>\w+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://bit.ly/{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        res = await session.get(
            f"https://bit.ly/{id}",
        )
        context.create_result("Bitly", url=url, score=1.0)
        context.visit(res.url)



import re
from typing import List, TypedDict

import aiohttp
from loguru import logger

from iwashi.helper import BASE_HEADERS, HTTP_REGEX
from iwashi.visitor import Context, SiteVisitor


class Instagram(SiteVisitor):
    NAME = "Instagram"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"instagram\.com/(?P<id>\w+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://www.instagram.com/{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        session = aiohttp.ClientSession(
            headers={
                "authority": "www.instagram.com",
                "accept": "*/*",
                "accept-language": "en-US,en;q=0.9",
                "referer": f"https://www.instagram.com/{id}/",
                "sec-ch-prefers-color-scheme": "dark",
                "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Microsoft Edge";v="108"',
                "sec-fetch-dest": "empty",
                "sec-fetch-mode": "cors",
                "sec-fetch-site": "same-origin",
                "x-asbd-id": "198387",
                "x-ig-www-claim": "0",
                "x-requested-with": "XMLHttpRequest",
            }
            | BASE_HEADERS,
        )

        url = f"https://www.instagram.com/{id}/"
        res = await session.get(url)
        match = re.search(r"\"X-IG-App-ID\": ?\"(?P<id>\d{15})\"", await res.text())
        if match is None:
            logger.warning(f"[Instagram] No X-IG-App-ID found in {url}")
            return
        session.headers["x-ig-app-id"] = match.group("id")

        csrf_res = await session.get("https://www.instagram.com/ajax/bz?__d=dis")
        session.headers["x-csrftoken"] = csrf_res.cookies["csrftoken"]

        info_res = await session.get(
            f"https://www.instagram.com/api/v1/users/web_profile_info/?username={id}",
        )
        if info_res.status // 100 != 2 or info_res.history:
            logger.warning("[Instagram] Blocked by Instagram")
            context.create_result(
                "Instagram",
                url=url,
                name=id,
                score=0.0,
                description="Blocked by Instagram",
            )
            return
        info: Root = await info_res.json()
        user = info["data"]["user"]
        context.create_result(
            "Instagram", url=url, score=1.0, description=user["biography"]
        )

        for link in user["bio_links"]:
            context.visit(link["url"])


class BioLinksItem0(TypedDict):
    title: str
    lynx_url: str
    url: str
    link_type: str


class BiographyWithEntities(TypedDict):
    raw_text: str
    entities: List


class EdgeFollowedBy(TypedDict):
    count: int


class EdgeMutualFollowedBy(TypedDict):
    count: int
    edges: List


class User(TypedDict):
    biography: str
    bio_links: List[BioLinksItem0]
    biography_with_entities: BiographyWithEntities
    blocked_by_viewer: bool
    restricted_by_viewer: None
    country_block: bool
    external_url: str
    external_url_linkshimmed: str
    edge_followed_by: EdgeFollowedBy
    fbid: str
    followed_by_viewer: bool
    edge_follow: EdgeFollowedBy
    follows_viewer: bool
    full_name: str
    group_metadata: None
    has_ar_effects: bool
    has_clips: bool
    has_guides: bool
    has_channel: bool
    has_blocked_viewer: bool
    highlight_reel_count: int
    has_requested_viewer: bool
    hide_like_and_view_counts: bool
    id: str
    is_business_account: bool
    is_professional_account: bool
    is_supervision_enabled: bool
    is_guardian_of_viewer: bool
    is_supervised_by_viewer: bool
    is_supervised_user: bool
    is_embeds_disabled: bool
    is_joined_recently: bool
    guardian_id: None
    business_address_json: None
    business_contact_method: str
    business_email: None
    business_phone_number: None
    business_category_name: None
    overall_category_name: None
    category_enum: None
    category_name: str
    is_private: bool
    is_verified: bool
    edge_mutual_followed_by: EdgeMutualFollowedBy
    profile_pic_url: str
    profile_pic_url_hd: str
    requested_by_viewer: bool
    should_show_category: bool
    should_show_public_contacts: bool
    transparency_label: None
    transparency_product: str
    username: str
    connected_fb_page: None
    pronouns: List


class Data(TypedDict):
    user: User


class Root(TypedDict):
    data: Data
    status: str



from __future__ import annotations

import json
import re
from typing import Dict, List, TypedDict

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Sketch(SiteVisitor):
    NAME = "Sketch"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"sketch\.pixiv\.net/@(?P<id>\w+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://sketch.pixiv.net/@{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        res = await session.get(
            f"https://sketch.pixiv.net/@{id}",
        )
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        # __NEXT_DATA__
        element: bs4.Tag = soup.find(attrs={"id": "__NEXT_DATA__"})  # type: ignore
        if element is None:
            logger.warning(f"__NEXT_DATA__ not found: {id}")
        next_data = json.loads(element.text)
        data: Root = json.loads(next_data["props"]["pageProps"]["initialState"])
        lives = tuple(data["live"]["lives"].values())
        if len(lives) == 0:
            return
        live = lives[0]
        users = data["users"]["users"]
        user_id = live["owner"]["user_id"]
        user = users[user_id]
        context.create_result(
            "Pixiv Sketch",
            url=url,
            name=user["name"],
            score=1.0,
            description=user["description"],
            profile_picture=user["icon"]["photo"]["original"]["url"],
        )

        context.visit(f'https://pixiv.net/users/{user["pixiv_user_id"]}')


class DescriptionFragmentsItem(TypedDict):
    type: str
    body: str
    normalized_body: str


class Sq800(TypedDict):
    url: str


class Thumbnail(TypedDict):
    sq800: Sq800
    w160: Sq800
    w400: Sq800
    w1280: Sq800
    original: Sq800


class Owner(TypedDict):
    user_id: str
    channel_id: str
    thumbnail: Thumbnail
    hls_movie: str
    is_enabled_mic_input: bool


class LiveInfo(TypedDict):
    finished_at: None
    description_fragments: List[DescriptionFragmentsItem]
    is_r15: bool
    is_closed: bool
    mode: str
    server: str
    is_r18: bool
    publicity: str
    is_adult: bool
    performers: List
    chat_count: int
    created_at: str
    member_count: int
    is_enabled_chat: bool
    name: str
    thumbnail: Thumbnail
    total_audience_count: int
    owner: Owner
    audience_count: int
    is_broadcasting: bool
    heart_count: int
    channel_id: str
    deleted: bool
    source: str
    id: str
    description: str
    is_single: bool
    performer_count: int
    is_enabled_mic_input: bool
    is_enabled_gifting: bool


class Options(TypedDict):
    filterBy: None
    orderBy: str


class Livelist(TypedDict):
    ids: List
    ended: bool
    fetching: bool
    options: Options


class Live(TypedDict):
    lives: Dict[str, LiveInfo]
    liveAdStages: Dict
    liveAvailability: bool
    liveClosedAvailability: bool
    liveLogs: Dict
    liveCaptions: Dict
    liveChatErrorMessage: None
    liveArchiveLink: Dict
    liveGiftings: Dict
    liveGiftingSummaries: Dict
    liveGiftingRecommends: Dict
    liveGiftingRecommendsPaging: Dict
    liveGiftingHistories: Dict
    liveSummaryUrl: None
    liveList: Livelist
    unpresentedBlacklistedLiveChats: List


class Color(TypedDict):
    hex: str
    r: int
    g: int
    b: int


class Pxsq60(TypedDict):
    width: int
    height: int
    url: str
    url2x: str


class Photo(TypedDict):
    pxsq60: Pxsq60
    pxw540: Pxsq60
    pxsq180: Pxsq60
    pxsq120: Pxsq60
    sq180: Pxsq60
    original: Pxsq60
    sq120: Pxsq60
    w240: Pxsq60
    sq60: Pxsq60
    w540: Pxsq60
    pxw240: Pxsq60


class Icon(TypedDict):
    id: int
    type: str
    color: Color
    photo: Photo


class Self(TypedDict):
    method: str
    href: str


class _Links(TypedDict):
    self: Self


class Stats(TypedDict):
    follower_count: int
    following_count: int
    heart_count: int
    resnap_count: int
    public_post_count: int


class UserInfo(TypedDict):
    description_fragments: List
    name: str
    followed: bool
    following: bool
    blocking: bool
    icon: Icon
    unique_name: str
    post_ids: List
    _links: _Links
    id: str
    pixiv_user_id: str
    description: str
    stats: Stats


Usercolors = TypedDict(
    "userColors",
    {"15241365": "str", "17391869": "str", "30087679": "str", "50150007": "str"},
)


class Users0(TypedDict):
    users: Dict[str, UserInfo]
    userWalls: Dict
    userLinks: Dict
    userColors: Usercolors
    usersRecommended: List
    usersBlockedWall: List
    usersBlockedWallLink: Dict


class Params(TypedDict):
    id: str


class Navigation(TypedDict):
    name: str
    params: Params
    search: Dict


class Status(TypedDict):
    code: str
    message: None


class Route(TypedDict):
    historyState: None
    isNavigating: bool
    navigation: Navigation
    status: Status


class Root(TypedDict):
    live: Live
    users: Users0
    route: Route



import re

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX, normalize_url, session
from iwashi.visitor import Context, SiteVisitor


class TwitCasting(SiteVisitor):
    NAME = "TwitCasting"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"twitcasting\.tv/(?P<id>[-\w]+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://twitcasting.tv/{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        url = f"https://twitcasting.tv/{id}"
        res = await session.get(
            url,
        )
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        # name: .tw-user-nav-name
        element = soup.select_one(".tw-user-nav-name")
        if element is None:
            logger.warning(f"[TwitCasting] Could not find name for {url}")
            return
        name = element.text.strip()
        # icon: .tw-user-nav-icon > img
        element = soup.select_one(".tw-user-nav-icon > img")
        profile_picture = None
        if element is not None:
            attr = element["src"]
            if isinstance(attr, str):
                profile_picture = normalize_url(attr)

        context.create_result(
            "TwitCasting",
            url=url,
            name=name,
            score=1.0,
            description=None,
            profile_picture=profile_picture,
        )

        links = set()
        for element in soup.select(".tw-follow-list-row-icon"):
            links.add(element["href"])
        for link in links:
            if link.startswith("/"):
                continue
            context.visit(link)



import re
from typing import List, TypedDict

from loguru import logger

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Fanbox(SiteVisitor):
    NAME = "Fanbox"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(?P<id>[\w-]+)\.fanbox\.cc", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://{match.group("id")}.fanbox.cc'

    async def visit(self, url, context: Context, id: str):
        url = f"https://{id}.fanbox.cc"
        creator_res = await session.get(
            f"https://api.fanbox.cc/creator.get?creatorId={id}",
            headers={
                "accept": "application/json",
                "origin": f"https://{id}.fanbox.cc",
                "referer": f"https://{id}.fanbox.cc/",
            },
        )
        if creator_res.status // 100 == 4:
            logger.warning(f"[Fanbox] Could not find user for {url}")
            return

        info: Root = await creator_res.json()
        if "error" in info:
            logger.warning(f"[Fanbox] Could not find user for {url}")
            return
        context.create_result(
            "Fanbox",
            url=url,
            name=info["body"]["user"]["name"],
            score=1.0,
            description=info["body"]["description"],
            profile_picture=info["body"]["user"]["iconUrl"],
        )

        for link in info["body"]["profileLinks"]:
            context.visit(link)


class User(TypedDict):
    userId: str
    name: str
    iconUrl: str


class ProfileItemsItem0(TypedDict):
    id: str
    type: str
    imageUrl: str
    thumbnailUrl: str


class Body(TypedDict):
    user: User
    creatorId: str
    description: str
    hasAdultContent: bool
    coverImageUrl: str
    profileLinks: List[str]
    profileItems: List[ProfileItemsItem0]
    isFollowed: bool
    isSupported: bool
    isStopped: bool
    isAcceptingRequest: bool
    hasBoothShop: bool


class Root(TypedDict):
    body: Body



from __future__ import annotations

import json
import re
from typing import List, TypedDict, Union

import bs4

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor

DATA_REGEX = r"preloadLink\s?=\s(?P<json>{[^;]+)"


class Fanlink(SiteVisitor):
    NAME = "Fanlink"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(?P<id>[\w-]+\.)?fanlink\.to/(?P<slug>[\w-]+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        id = match.group("id")
        if id is None:
            return f'https://fanlink.to/{match.group("slug")}'
        return f'https://{id}fanlink.to/{match.group("slug")}'

    async def visit(
        self, url, context: Context, id: str | None = None, slug: str | None = None
    ) -> None:
        res = await session.get(url)
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        for script in soup.find_all("script"):
            if not script.string:
                continue
            match = re.search(DATA_REGEX, script.string)
            if match is None:
                continue
            data = json.loads(match.group("json"))
            break
        else:
            raise Exception("No script found")

        context.create_result(
            site_name="Fanlink",
            url=url,
            name=data["title"],
            description=data["description"],
            profile_picture=data["image_url"],
        )

        for link in data["social_settings"]:
            context.visit(link["url"])


class ScheduleItem(TypedDict):
    type: str
    start_date: str


class PixelsItem(TypedDict):
    platform: str


class ServicesItem(TypedDict):
    id: int
    url: str
    type: str
    active: bool
    track_id: str
    service_name: str
    match_confidence: int
    clickthrough_count: int


class ServicesItem0(TypedDict):
    id: int
    url: str
    isrc: str
    type: str
    active: bool
    track_id: str
    preview_url: str
    service_name: str
    match_confidence: int
    clickthrough_count: int


class ServicesItem1(TypedDict):
    id: int
    url: str
    type: str
    active: bool
    track_id: str
    preview_url: str
    service_name: str
    match_confidence: int
    clickthrough_count: int


class ServicesItem2(TypedDict):
    id: int
    active: bool
    rescan: bool
    service_name: str
    match_confidence: None
    clickthrough_count: int


class ServicesItem3(TypedDict):
    id: int
    url: str
    active: bool
    track_id: str
    preview_url: str
    service_name: str
    match_confidence: int
    clickthrough_count: int


class ServicesItem4(TypedDict):
    id: int
    active: bool
    service_name: str
    attachment_id: int
    clickthrough_count: int


class SocialSettingsItem(TypedDict):
    url: str
    color: str
    active: bool
    platform: str


class Root(TypedDict):
    click_count: int
    past_events: None
    schedule: List[ScheduleItem]
    id: int
    active_job_id: None
    author: str
    call_to_action: str
    button_color: str
    bg_color: None
    clickthrough_count: int
    emails_sent: None
    custom_domain: None
    date: None
    description: None
    display_content: None
    location: None
    price: None
    image_url: str
    bg_url: None
    is_hidden: bool
    is_active: bool
    message_settings: None
    isrc: None
    upc: None
    metadata_description: str
    metadata_image_url: str
    metadata_title: str
    meta_tags: List
    preview_url: str
    autoplay_preview: bool
    rsvp_attachment_id: None
    mock_insights: bool
    sort_services_by_clickthroughs: bool
    affiliate_codes: None
    pixels: List[PixelsItem]
    rules: List
    services: List[
        Union[
            ServicesItem0,
            ServicesItem3,
            ServicesItem,
            ServicesItem1,
            ServicesItem4,
            ServicesItem2,
        ]
    ]
    shortened_path: str
    show_socials: bool
    skip_landing_page: bool
    social_settings: List[SocialSettingsItem]
    subdomain: str
    target_type: str
    target_url: str
    title: str
    header: str
    use_saved_service: bool
    user_id: int
    createdAt: str
    updatedAt: str
    rsvp_attachment: None



import json
import re
from typing import List, TypedDict

from loguru import logger

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Soundcloud(SiteVisitor):
    NAME = "Soundcloud"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"soundcloud\.com/(?P<id>[-\w]+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://soundcloud.com/{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        url = f"https://soundcloud.com/{id}"
        res = await session.get(
            url,
        )
        info_json = re.search(
            r"window\.__sc_hydration ?= ?(?P<info>.+);", await res.text()
        )
        if info_json is None:
            logger.warning(f"[Soundcloud] Could not find info for {url}")
            return
        info_list: Root = json.loads(info_json.group("info"))
        for info in info_list:
            if info["hydratable"] == "user":
                break
        else:
            logger.warning(f"[Soundcloud] Could not find user info for {url}")
            return

        context.create_result(
            "Soundcloud",
            url=url,
            name=info["data"]["username"],
            score=1.0,
            description=info["data"]["description"],
            profile_picture=info["data"]["avatar_url"],
        )

        client_id_res = await session.get(
            "https://a-v2.sndcdn.com/assets/0-bf97f26a.js",
        )
        match = re.search(
            r"client_id: ?\"(?P<client_id>\w{32})\"", await client_id_res.text()
        )
        if match is None:
            logger.warning(f"[Soundcloud] Could not find client_id for {url}")
            return
        client_id = match.group("client_id")

        headers = {
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "Accept-Language": "ja,en;q=0.9,en-GB;q=0.8,en-US;q=0.7",
            "Authorization": "OAuth 2-294132-999490816-Mf51UTFjQ256LP",
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
            "Origin": "https://soundcloud.com",
            "Pragma": "no-cache",
            "Referer": "https://soundcloud.com/",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-site",
            "sec-ch-ua": '"Chromium";v="118", "Microsoft Edge";v="118", "Not=A?Brand";v="99"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
        }

        params = {
            "client_id": client_id,
            "app_version": "1697797929",
            "app_locale": "en",
        }

        profile_res = await session.get(
            f'https://api-v2.soundcloud.com/users/{info["data"]["urn"]}/web-profiles',
            params=params,
            headers=headers,
        )
        profile: List[ProfileItem] = await profile_res.json()
        for item in profile:
            context.visit(item["url"])


class ProfileItem(TypedDict):
    url: str
    network: str
    title: str


class Product(TypedDict):
    id: str


class CreatorSubscriptionsItem0(TypedDict):
    product: Product


class VisualsItem0(TypedDict):
    urn: str
    entry_time: int
    visual_url: str


class Visuals(TypedDict):
    urn: str
    enabled: bool
    visuals: List[VisualsItem0]
    tracking: None


class Badges(TypedDict):
    pro: bool
    pro_unlimited: bool
    verified: bool


class Data4(TypedDict):
    avatar_url: str
    city: str
    comments_count: int
    country_code: str
    created_at: str
    creator_subscriptions: List[CreatorSubscriptionsItem0]
    creator_subscription: CreatorSubscriptionsItem0
    description: str
    followers_count: int
    followings_count: int
    first_name: str
    full_name: str
    groups_count: int
    id: int
    kind: str
    last_modified: str
    last_name: str
    likes_count: int
    playlist_likes_count: int
    permalink: str
    permalink_url: str
    playlist_count: int
    reposts_count: None
    track_count: int
    uri: str
    urn: str
    username: str
    verified: bool
    visuals: Visuals
    badges: Badges
    station_urn: str
    station_permalink: str
    url: str


class RootItem0(TypedDict):
    hydratable: str
    data: Data4


Root = List[RootItem0]



from __future__ import annotations

import json
import re
from typing import TypedDict

import bs4

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Note(SiteVisitor):
    NAME = "Note"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"note\.com/(?P<user>[^/]+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f"https://note.com/{match.group('user')}"

    async def visit(self, url: str, context: Context, user: str):
        res = await session.get(
            url,
        )
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")

        data_element = soup.select_one("script[type='application/ld+json']")
        if data_element is None:
            return
        data: Root = json.loads(data_element.text)

        context.create_result(
            "Note",
            url=url,
            score=1.0,
            name=data["headline"],
            description=data["description"],
            profile_picture=data["image"]["url"],
        )

        links: set[str] = set()
        for element in soup.select(".m-creatorSocialLinks__item"):
            link = element.select_one("a")
            if link is None:
                continue
            if "href" not in link.attrs:
                continue
            links.add(link.attrs["href"])
        for link in links:
            context.visit(link)


Author = TypedDict("author", {"@type": "str", "name": "str", "url": "str"})
Logo = TypedDict(
    "logo", {"@type": "str", "url": "str", "width": "str", "height": "str"}
)
Publisher = TypedDict("publisher", {"@type": "str", "name": "str", "logo": "Logo"})
Image = TypedDict(
    "image", {"@type": "str", "url": "str", "width": "int", "height": "int"}
)
Root = TypedDict(
    "Root",
    {
        "@context": "str",
        "@type": "str",
        "mainEntityOfPage": "str",
        "headline": "str",
        "datePublished": "str",
        "dateModified": "str",
        "author": "Author",
        "publisher": "Publisher",
        "image": "Image",
        "description": "str",
    },
)



from __future__ import annotations

import abc
import re
from dataclasses import dataclass, field
from typing import List, Optional, Protocol

HTTP_REGEX = "(https?://)?(www.)?"


@dataclass
class Result:
    url: str
    score: float
    site_name: Optional[str]
    title: Optional[str]
    description: Optional[str]
    profile_picture: Optional[str]

    children: List["Result"] = field(default_factory=list)
    links: List["str"] = field(default_factory=list)

    def to_list(self) -> List[Result]:
        links: List[Result] = [self]
        for child in self.children:
            links.extend(child.to_list())
        return links


class Visitor(Protocol):
    async def visit(self, url: str, context: "Context", **kwargs) -> Result:
        ...

    async def tree(self, url: str, context: "Context", **kwargs) -> Result:
        ...

    def queue(self, url: str, context: "Context") -> None:
        ...

    def mark_visited(self, url: str) -> None:
        ...


class Context:
    def __init__(self, url: str, visitor: Visitor, parent: Optional["Context"] = None):
        self.url = url
        self.visitor = visitor
        self.parent = parent
        self.result: Optional[Result] = None

    def create_result(
        self,
        site_name: str,
        url: str,
        score: float = 1.0,
        name: Optional[str] = None,
        description: Optional[str] = None,
        profile_picture: Optional[str] = None,
    ) -> Result:
        self.result = Result(
            site_name=site_name,
            url=url,
            score=score,
            title=name,
            description=description,
            profile_picture=profile_picture,
        )

        if self.parent and self.parent.result:
            self.parent.result.children.append(self.result)

        return self.result

    def link(self, url: str) -> None:
        if self.result is not None:
            self.result.links.append(url)

    def mark_visited(self, url: str) -> None:
        self.visitor.mark_visited(url)

    def create(self, url: str) -> "Context":
        return Context(url=url, visitor=self.visitor, parent=self)

    def visit(self, url: str) -> None:
        self.visitor.queue(url, self.create(url))

    def __repr__(self) -> str:
        return f"Context(url={self.url!r})"


class SiteVisitor(abc.ABC):
    NAME: Optional[str] = None
    URL_REGEX: Optional[re.Pattern] = None

    def match(self, url, context: Context) -> Optional[re.Match]:
        if self.URL_REGEX is None:
            raise NotImplementedError()
        return self.URL_REGEX.match(url)

    async def normalize(self, url: str) -> str | None:
        return url

    @abc.abstractmethod
    async def visit(self, url, context: Context, **kwargs) -> Optional[Result]:
        raise NotImplementedError()
        raise NotImplementedError()



from __future__ import annotations

import re
from typing import List, TypedDict

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Misskey(SiteVisitor):
    NAME = "Misskey"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(?P<host>[^.]+\.[^\/]+)/channels/(?P<channek_id>[\w]+)",
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://{match.group("host")}/channels/{match.group("channel_id")}'

    async def visit(self, url, context: Context, host: str, channel_id: str):
        url = f"https://{host}/api/channels/show"
        res = await session.post(
            url,
            json={"channelId": channel_id},
        )
        info: Root = await res.json()
        context.create_result(
            "Misskey",
            url=f"https://{host}/channels/{channel_id}",
            name=info["name"],
            score=1.0,
            description=info["description"],
            profile_picture=info["bannerUrl"],
        )


class Root(TypedDict):
    id: str
    createdAt: str
    lastNotedAt: str
    name: str
    description: str
    userId: str
    bannerUrl: str
    pinnedNoteIds: List
    color: str
    isArchived: bool
    usersCount: int
    notesCount: int
    isFollowing: bool
    isFavorited: bool
    hasUnreadNote: bool
    pinnedNotes: List



import asyncio
from typing import List, MutableSet, Optional

import aiohttp
from loguru import logger

from .helper import BASE_HEADERS, DEBUG, normalize_url, parse_host, session
from .visitor import Context, Result, SiteVisitor, Visitor


class Iwashi(Visitor):
    def __init__(self) -> None:
        self.visitors: List[SiteVisitor] = []
        self.visited: MutableSet[str] = set()
        self.tasks: List[asyncio.Task] = []

    def add_visitor(self, visitor: SiteVisitor) -> None:
        self.visitors.append(visitor)

    def is_visited(self, url: str) -> bool:
        return url in self.visited

    def mark_visited(self, url: str) -> bool:
        url = url.lower()
        if self.is_visited(url):
            return False
        self.visited.add(url)
        return True

    def queue(self, url: str, context: Context) -> None:
        loop = asyncio.get_event_loop()
        self.tasks.append(loop.create_task(self.visit(url, context)))

    async def tree(self, url: str, context: Optional[Context] = None) -> Result | None:
        context = context or Context(url=url, visitor=self)
        context = context.create(url)
        await self.visit(url, context)
        while self.tasks:
            task = self.tasks.pop(0)
            await task
        return context.result

    async def visit(
        self, _url: str, context: Optional[Context] = None
    ) -> Optional[Result]:
        url = normalize_url(_url)
        if url is None:
            return None
        context = context or Context(url=url, visitor=self)
        if self.is_visited(url):
            return None
        for visitor in self.visitors:
            match = visitor.match(url, context)
            if match is None:
                continue

            try:
                normalized = await visitor.normalize(url)
                if normalized is None:
                    continue
                if self.mark_visited(normalized):
                    match = visitor.match(normalized, context)
                    if match is not None:
                        await visitor.visit(normalized, context, **match.groupdict())
                elif context.parent is not None:
                    context.parent.link(normalized)
            except Exception as e:
                logger.warning(f"[Visitor Error] {url} {visitor.__class__.__name__}")
                logger.exception(e)
                if DEBUG:
                    raise e
                continue
            break
        else:
            context.create_result(site_name=parse_host(url), url=url, score=1.0)
            self.mark_visited(url)
            if await self.try_redirect(url, context):
                return context.result
            else:
                logger.warning(f"[No Visitor Found] {url}")

        return context.result

    async def try_redirect(self, url: str, context: Context) -> bool:
        try:
            res = await session.get(url, headers=BASE_HEADERS, allow_redirects=True)
        except aiohttp.ClientError:
            logger.warning(f"[Redirect] failed to redirect {url}")
            return False
        new_url = str(res.url)
        if new_url == url:
            return False
        context.create_result(site_name=parse_host(url), url=url, score=1.0)
        context.visit(new_url)
        logger.info(f"[Redirect] {url} -> {new_url}")
        return True


def get_iwashi():
    iwashi = Iwashi()
    add_visitors(iwashi)
    return iwashi


def add_visitors(iwashi):
    from . import visitors

    for attr in dir(visitors):
        value = getattr(visitors, attr)
        if attr.startswith("_"):
            continue
        if isinstance(value, type) and issubclass(value, SiteVisitor):
            iwashi.add_visitor(value())


async def visit(url: str, iwashi: Optional[Iwashi] = None) -> Optional[Result]:
    iwashi = iwashi or get_iwashi()
    return await iwashi.tree(url)



from typing import List, NotRequired, TypedDict


class thumbnail(TypedDict):
    url: str
    width: int
    height: int


class thumbnails(TypedDict):
    thumbnails: List[thumbnail]


class urlEndpoint(TypedDict):
    url: str


class webCommandMetadata(TypedDict):
    sendPost: bool
    apiUrl: str


class commandMetadata(TypedDict):
    webCommandMetadata: webCommandMetadata
    sendPost: bool
    apiUrl: str


class continuationCommand(TypedDict):
    token: str
    request: str


class continuationEndpoint(TypedDict):
    commandMetadata: commandMetadata
    continuationCommand: continuationCommand


class continuationItemRenderer(TypedDict):
    continuationEndpoint: continuationEndpoint


class itemSectionRenderer_content(TypedDict):
    continuationItemRenderer: continuationItemRenderer


class itemSectionRenderer(TypedDict):
    contents: List[itemSectionRenderer_content]


class sectionListRenderer_contents(TypedDict):
    itemSectionRenderer: itemSectionRenderer


class sectionListRenderer(TypedDict):
    contents: List[sectionListRenderer_contents]


class content(TypedDict):
    sectionListRenderer: sectionListRenderer


class engagementPanelSectionListRenderer(TypedDict):
    content: content


class engagementPanel(TypedDict):
    engagementPanelSectionListRenderer: engagementPanelSectionListRenderer


class showEngagementPanelEndpoint(TypedDict):
    engagementPanel: engagementPanel


class innertubeCommand(TypedDict):
    urlEndpoint: urlEndpoint
    showEngagementPanelEndpoint: showEngagementPanelEndpoint


class onTap(TypedDict):
    innertubeCommand: innertubeCommand


class _commandRuns(TypedDict):
    onTap: onTap


class firstLink(TypedDict):
    content: str
    commandRuns: List[_commandRuns]


class more(TypedDict):
    content: str
    commandRuns: List[_commandRuns]


class channelHeaderLinksViewModel(TypedDict):
    firstLink: NotRequired[firstLink]
    more: NotRequired[more]


class headerLinks(TypedDict):
    channelHeaderLinksViewModel: channelHeaderLinksViewModel


class runs_item(TypedDict):
    text: str


class runs(TypedDict):
    runs: List[runs_item]


class c4TabbedHeaderRenderer(TypedDict):
    channelId: str
    title: str
    avatar: thumbnails
    banner: thumbnails
    headerLinks: NotRequired[headerLinks]
    trackingParams: str
    channelHandleText: runs
    style: str


class header(TypedDict):
    c4TabbedHeaderRenderer: c4TabbedHeaderRenderer


class channelMetadataRenderer(TypedDict):
    title: str
    description: str
    rssUrl: str
    externalId: str
    keywords: str
    ownerUrls: List[str]
    avatar: thumbnails
    channelUrl: str
    isFamilySafe: bool
    vanityChannelUrl: str


class metadata(TypedDict):
    channelMetadataRenderer: channelMetadataRenderer


class link(TypedDict):
    hrefUrl: str


class linkAlternates(TypedDict):
    linkAlternates: List[link]


class microformatDataRenderer(TypedDict):
    urlCanonical: str
    title: str
    description: str
    thumbnail: thumbnails
    siteName: str
    appName: str
    urlApplinksWeb: str
    schemaDotOrgType: str
    noindex: bool
    unlisted: bool
    familySafe: bool
    linkAlternates: linkAlternates


class microformat(TypedDict):
    microformatDataRenderer: microformatDataRenderer


class ytinitialdata(TypedDict):
    header: header
    metadata: metadata
    microformat: microformat



from __future__ import annotations

import json
import re
from typing import List, TypedDict

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX, normalize_url, session
from iwashi.visitor import Context, SiteVisitor


class TikTok(SiteVisitor):
    NAME = "TikTok"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"tiktok\.com/@(?P<id>[-\w]+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://tiktok.com/@{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        url = f"https://tiktok.com/@{id}"
        res = await session.get(
            url,
        )
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        # icon: <meta property="og:image"
        element = soup.select_one('meta[property="og:image"]')
        profile_picture = None
        if element is not None:
            attr = element["content"]
            if isinstance(attr, str):
                profile_picture = normalize_url(attr)

        # data: #Person
        element = soup.select_one("script#Person")
        if element is None:
            logger.warning(f"[TikTok] Could not find data for {url}")
            return
        data: Root = json.loads(element.text)

        context.create_result(
            "TikTok",
            url=url,
            name=data["name"],
            score=1.0,
            description=data["description"],
            profile_picture=profile_picture,
        )


Interactiontype = TypedDict("interactionType", {"@type": "str"})
InteractionstatisticItem = TypedDict(
    "interactionStatistic_item",
    {
        "@type": "str",
        "interactionType": "Interactiontype",
        "userInteractionCount": "int",
    },
)
Mainentityofpage = TypedDict("mainEntityOfPage", {"@id": "str", "@type": "str"})
Root = TypedDict(
    "Root",
    {
        "@context": "str",
        "@type": "str",
        "name": "str",
        "description": "str",
        "alternateName": "str",
        "url": "str",
        "knowsLanguage": "str",
        "nationality": "str",
        "interactionStatistic": "List[InteractionstatisticItem]",
        "mainEntityOfPage": "Mainentityofpage",
    },
)



from __future__ import annotations

from typing import List, TypedDict


class Title(TypedDict):
    content: str


class Channelexternallinkviewmodel(TypedDict):
    title: Title
    link: Title


class LinksItem(TypedDict):
    channelExternalLinkViewModel: Channelexternallinkviewmodel


class Aboutchannelviewmodel(TypedDict):
    description: str
    links: List[LinksItem]


class Metadata(TypedDict):
    aboutChannelViewModel: Aboutchannelviewmodel


class Aboutchannelrenderer(TypedDict):
    metadata: Metadata


class ContinuationitemsItem(TypedDict):
    aboutChannelRenderer: Aboutchannelrenderer


class Appendcontinuationitemsaction(TypedDict):
    continuationItems: List[ContinuationitemsItem]


class OnresponsereceivedendpointsItem(TypedDict):
    appendContinuationItemsAction: Appendcontinuationitemsaction


class AboutRes(TypedDict):
    onResponseReceivedEndpoints: List[OnresponsereceivedendpointsItem]



from __future__ import annotations

import re
import urllib.parse

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, SiteVisitor


class PageLink(SiteVisitor):
    NAME = "PageLink"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(?P<host>\w+)\.page\.link/\?link=(?P<link>[^&]+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://{match.group("host")}.page.link/?link={match.group("link")}'

    async def visit(self, url, context: Context, host: str, link: str):
        context.create_result(
            "PageLink",
            url=url,
            score=1.0,
        )
        context.visit(urllib.parse.unquote(link))



from __future__ import annotations

import json
import re
from typing import List, TypedDict, Union

import bs4

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor

DATA_REGEX = r"preloadLink\s?=\s(?P<json>{[^;]+)"


class Bandcamp(SiteVisitor):
    NAME = "Bandcamp"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(?P<id>[\w-]+\.)?bandcamp\.com/(?P<slug>[\w-]+)?", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return (
            f'https://{match.group("id") or ""}bandcamp.com/{match.group("slug") or ""}'
        )

    async def visit(
        self, url, context: Context, id: str | None = None, slug: str | None = None
    ) -> None:
        res = await session.get(url)
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        script = soup.select_one("script[type='application/ld+json']")
        if script is None:
            raise Exception("No script found")

        data: Root = json.loads(script.text)
        context.create_result(
            site_name="Bandcamp",
            url=url,
            name=data["name"],
            description=data.get("description"),
            profile_picture=data["image"],
        )

        for link in data["publisher"]["mainEntityOfPage"]:
            if link["@type"] != "WebPage":
                continue
            context.visit(link["url"])


AdditionalpropertyItem = TypedDict(
    "additionalProperty_item", {"@type": "str", "name": "str", "value": "int"}
)
AdditionalpropertyItem0 = TypedDict(
    "additionalProperty_item", {"@type": "str", "name": "str", "value": "str"}
)


class Pricespecification(TypedDict):
    minPrice: float


Offers = TypedDict(
    "offers",
    {
        "@type": "str",
        "url": "str",
        "priceCurrency": "str",
        "price": "float",
        "priceSpecification": "Pricespecification",
        "availability": "str",
    },
)
AlbumreleaseItem = TypedDict(
    "albumRelease_item",
    {
        "@type": "List[str]",
        "@id": "str",
        "name": "str",
        "additionalProperty": "List[Union[AdditionalpropertyItem, AdditionalpropertyItem0]]",
        "description": "str",
        "offers": "Offers",
        "musicReleaseFormat": "str",
        "image": "List[str]",
    },
)
AdditionalpropertyItem1 = TypedDict(
    "additionalProperty_item", {"@type": "str", "name": "str", "value": "bool"}
)
AdditionalpropertyItem2 = TypedDict(
    "additionalProperty_item", {"@type": "str", "name": "str", "value": "float"}
)
Offers0 = TypedDict(
    "offers",
    {
        "@type": "str",
        "url": "str",
        "priceCurrency": "str",
        "price": "float",
        "priceSpecification": "Pricespecification",
        "availability": "str",
        "additionalProperty": "List[Union[AdditionalpropertyItem, AdditionalpropertyItem2]]",
    },
)
AlbumreleaseItem0 = TypedDict(
    "albumRelease_item",
    {
        "@type": "List[str]",
        "@id": "str",
        "name": "str",
        "additionalProperty": "List[Union[AdditionalpropertyItem, AdditionalpropertyItem1, AdditionalpropertyItem0]]",
        "offers": "Offers0",
        "musicReleaseFormat": "str",
        "image": "List[str]",
    },
)
Inalbum = TypedDict(
    "inAlbum",
    {
        "@type": "str",
        "name": "str",
        "albumRelease": "List[Union[AlbumreleaseItem0, AlbumreleaseItem]]",
        "albumReleaseType": "str",
        "numTracks": "int",
    },
)
Byartist = TypedDict("byArtist", {"@type": "str", "name": "str", "@id": "str"})
MainentityofpageItem = TypedDict(
    "mainEntityOfPage_item", {"@type": "str", "url": "str", "name": "str"}
)
SubjectofItem = TypedDict(
    "subjectOf_item",
    {
        "@type": "str",
        "url": "str",
        "name": "str",
        "additionalProperty": "List[AdditionalpropertyItem0]",
    },
)
Foundinglocation = TypedDict("foundingLocation", {"@type": "str", "name": "str"})
Publisher = TypedDict(
    "publisher",
    {
        "@type": "str",
        "@id": "str",
        "name": "str",
        "additionalProperty": "List[Union[AdditionalpropertyItem, AdditionalpropertyItem1]]",
        "image": "str",
        "genre": "str",
        "description": "str",
        "mainEntityOfPage": "List[MainentityofpageItem]",
        "subjectOf": "List[SubjectofItem]",
        "foundingLocation": "Foundinglocation",
    },
)
Author = TypedDict(
    "author",
    {
        "@type": "str",
        "url": "str",
        "image": "str",
        "additionalProperty": "List[AdditionalpropertyItem]",
        "name": "str",
    },
)
CommentItem = TypedDict(
    "comment_item", {"@type": "str", "author": "Author", "text": "List[str]"}
)
Root = TypedDict(
    "Root",
    {
        "@type": "str",
        "@id": "str",
        "additionalProperty": "List[Union[AdditionalpropertyItem, AdditionalpropertyItem0]]",
        "name": "str",
        "description": "str",
        "duration": "str",
        "dateModified": "str",
        "datePublished": "str",
        "inAlbum": "Inalbum",
        "byArtist": "Byartist",
        "publisher": "Publisher",
        "copyrightNotice": "str",
        "comment": "List[CommentItem]",
        "keywords": "List[str]",
        "image": "str",
        "sponsor": "List[Author]",
        "mainEntityOfPage": "str",
        "@context": "str",
    },
)



from .bandcamp import Bandcamp
from .bitly import Bitly
from .booth import Booth
from .fanbox import Fanbox
from .fanlink import Fanlink
from .googl import Googl
from .instagram import Instagram
from .linktree import Linktree
from .litlink import LitLink
from .mirrativ import Mirrativ
from .misskey import Misskey
from .nicovideo import Nicovideo
from .note import Note
from .pagelink import PageLink
from .pixiv import Pixiv
from .reddit import Reddit
from .sketch import Sketch
from .soundcloud import Soundcloud
from .tiktok import TikTok
from .twitcasting import TwitCasting
from .twitch import Twitch
from .twitter import Twitter
from .youtube import Youtube

__all__ = [
    "Bandcamp",
    "Bitly",
    "Booth",
    "Fanbox",
    "Fanlink",
    "Googl",
    "Instagram",
    "Linktree",
    "LitLink",
    "Mirrativ",
    "Misskey",
    "Nicovideo",
    "Note",
    "PageLink",
    "Pixiv",
    "Reddit",
    "Sketch",
    "Soundcloud",
    "TikTok",
    "TwitCasting",
    "Twitch",
    "Twitter",
    "Youtube",
]



import json
import re
from typing import List, Optional, TypedDict, Union

import bs4

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Linktree(SiteVisitor):
    NAME = "Linktree"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"linktr\.ee/(?P<id>\w+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://linktr.ee/{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        res = await session.get(
            f"https://linktr.ee/{id}",
        )
        if res.status != 200:
            return None
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        data_element = soup.find(attrs={"id": "__NEXT_DATA__"})
        assert data_element
        data: Root = json.loads(data_element.get_text())
        info = data["props"]["pageProps"]
        context.create_result(
            "Linktree",
            url=url,
            score=1.0,
            description=info["description"],
            profile_picture=info["profilePictureUrl"],
        )
        links = info["links"]
        for link in links:
            if not link["url"]:
                continue
            context.visit(link["url"])


class Owner(TypedDict):
    id: int
    isEmailVerified: bool


class Modifiers(TypedDict):
    animation: None
    isForwarding: bool
    isForwardingActive: bool
    thumbnailUrl: None
    amazonAffiliate: None


class Gate(TypedDict):
    activeOrder: List
    age: None
    passcode: None
    nft: None
    payment: None


class Rules(TypedDict):
    gate: Gate


class Context1(TypedDict):
    layoutOption: str


class Context2(TypedDict):
    channelId: str
    subscribe: None
    embedOption: str


class LinksItem0(TypedDict):
    id: int
    type: str
    title: str
    position: int
    url: str
    shouldRouteToProfile: bool
    modifiers: Modifiers
    context: Union[Context, Context1, Context2]
    rules: Rules


class Background(TypedDict):
    color: str
    style: str
    type: str


class BackgroundStyle(TypedDict):
    color: str


class ButtonStyle(TypedDict):
    type: str
    backgroundStyle: BackgroundStyle
    shadowStyle: BackgroundStyle
    textStyle: BackgroundStyle


class Typeface(TypedDict):
    color: str
    family: str


class Theme(TypedDict):
    key: str
    luminance: str
    background: Background
    buttonStyle: ButtonStyle
    socialStyle: BackgroundStyle
    typeface: Typeface


class ThemeV2(TypedDict):
    key: str
    luminance: str
    background: Background
    buttonStyle: ButtonStyle
    typeface: Typeface


class Account(TypedDict):
    id: int
    uuid: str
    username: str
    tier: str
    isActive: bool
    profilePictureUrl: str
    pageTitle: str
    googleAnalyticsId: None
    facebookPixelId: None
    tiktokPixelId: None
    donationsActive: bool
    causeBanner: None
    contentWarning: None
    description: str
    isLogoVisible: bool
    socialLinksPosition: str
    useFooterSignup: bool
    useSignupLink: bool
    createdAt: int
    updatedAt: int
    expandableLinkCaret: bool
    verticals: List[Optional[str]]
    customAvatar: str
    customAvatarAttributes: None
    backgroundImageAttributes: None
    showRepeatVisitorSignupCta: bool
    profileBadges: None
    isVenmoEnabled: bool
    isSquareWalletEnabled: bool
    isCookieBannerEnabled: bool
    isInitialsProfileEnabled: bool
    isWhatsappNotificationsEnabled: bool
    isShareLinksEnabled: bool
    isOnlyfansSEOEnabled: bool
    linkTypesForSEO: None
    manualTitleTag: None
    profileDirectoryUrl: None
    enableDynamicProfilePageMetadata: bool
    linkPlatforms: List[str]
    activeGates: List
    isAmazonAffiliateEnabled: bool
    profileLinkContentDisplayType: str
    complementaryThemeProperties: bool
    timezone: str
    affiliateTokens: List
    owner: Owner
    pageMeta: None
    integrations: List
    links: List[LinksItem0]
    socialLinks: List
    theme: Theme
    themeV2: ThemeV2


class SeoSchemaClassifications(TypedDict):
    typeClassification: None


class LinksItem01(TypedDict):
    id: str
    title: str
    context: Union[Context, Context2]
    animation: None
    thumbnail: None
    url: str
    amazonAffiliate: None
    type: str
    rules: Rules
    position: int
    locked: None


class ChildLinksItem0(TypedDict):
    id: str
    title: str
    context: Context2
    animation: None
    thumbnail: None
    url: str
    amazonAffiliate: None
    type: str
    rules: Rules
    position: int
    locked: None
    autoOpenOnActive: bool
    isContentMatchingType: bool


class LinksItem1(TypedDict):
    id: str
    title: str
    context: Context1
    animation: None
    thumbnail: None
    url: str
    amazonAffiliate: None
    type: str
    rules: Rules
    position: int
    locked: None
    childLinks: Union[List[ChildLinksItem0], List[ChildLinksItem0]]


class Environment(TypedDict):
    LINK_TYPES_ASSETS_ENDPOINT: str
    STRIPE_PAYMENTS_API_ENDPOINT: str
    STRIPE_PUBLISHABLE_KEY: str
    PAYPAL_PAYMENTS_API_ENDPOINT: str
    PAYPAL_PAYMENTS_CLIENT_ID: str
    SHOPIFY_INTEGRATIONS_API_ENDPOINT: str
    META_IMAGE_URL: str
    RECAPTCHA_SITE_KEY: str
    RECAPTCHA_SITE_KEY_INVISIBLE: str
    GRAPHQL_API_ENDPOINT: str
    PROFILES_API_HOST: str


class Auth0Config(TypedDict):
    clientID: str
    domain: str
    redirectUri: str
    responseType: str
    responseMode: str
    audience: str


class PageProps(TypedDict):
    account: Account
    theme: ThemeV2
    isProfileVerified: bool
    hasConsentedToView: bool
    username: str
    pageTitle: str
    description: str
    socialLinks: List
    integrations: List
    seoSchemaClassifications: SeoSchemaClassifications
    metaTitle: str
    metaDescription: str
    profilePictureUrl: str
    links: List[Union[LinksItem01, LinksItem1]]
    leapLink: None
    isOwner: bool
    isLogoVisible: bool
    mobileDetected: bool
    userAgent: str
    stage: str
    environment: Environment
    contentGating: str
    videoStructuredData: List
    hasSensitiveContent: bool
    auth0Config: Auth0Config
    followerNotificationsEnabled: bool
    followerCapabilities: None


class Props(TypedDict):
    pageProps: PageProps
    __N_SSP: bool


class Query(TypedDict):
    profile: str


class RuntimeConfig(TypedDict):
    INGRESS_API_ENDPOINT: str
    PAYMENTS_API_ENDPOINT: str
    ANALYTICS_SCRIPT_URL: str
    LINK_TYPES_ASSETS_ENDPOINT: str
    STAGE: str
    SERVICE: str
    WEB_VITALS: str
    DD_SAMPLE_RATE: str
    DD_CLIENT_TOKEN: str
    STRIPE_PAYMENTS_API_ENDPOINT: str
    STRIPE_PUBLISHABLE_KEY: str
    PAYPAL_PAYMENTS_API_ENDPOINT: str
    PAYPAL_PAYMENTS_CLIENT_ID: str
    SHOPIFY_INTEGRATIONS_API_ENDPOINT: str
    RECAPTCHA_SITE_KEY: str
    RECAPTCHA_SITE_KEY_INVISIBLE: str
    BASE_URL: str
    BASE_PROFILE_URL: str


class Root(TypedDict):
    props: Props
    page: str
    query: Query
    buildId: str
    assetPrefix: str
    runtimeConfig: RuntimeConfig
    isFallback: bool
    gssp: bool
    customServer: bool
    locale: str
    locales: List[str]
    defaultLocale: str
    scriptLoader: List



import asyncio
import random
import re
import string
import time
from typing import Any, Callable

import aiohttp
from loguru import logger

from .visitor import Result

USER_AGENT = "Profile Link Generator (https://github.com/am230/iwashi)"
BASE_HEADERS = {"User-Agent": USER_AGENT}
HTTP_REGEX = "(https?://)?(www.)?"
DEBUG = False
session = aiohttp.ClientSession(headers=BASE_HEADERS)


def print_result(
    result: Result, indent_level=0, print: Callable[[str], Any] = print
) -> None:
    indent = indent_level * "    "
    print(f"{indent}{result.site_name}")
    print(f"{indent}│url  : {result.url}")
    print(f"{indent}│name : {result.title}")
    print(f"{indent}│score: {result.score}")
    print(f"{indent}│links : {result.links}")
    if result.description:
        print(f"{indent}│description: " + result.description.replace("\n", "\\n"))
    for child in result.children:
        print_result(child, indent_level + 1, print)


def parse_host(url: str) -> str:
    match = re.search(r"(https?:\/\/)?(www\.)?(?P<host>[\w.]+)/", url)
    if match is None:
        return url
    return match.group("host")


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


URL_NORMALIZE_REGEX = r"(?P<protocol>https?)?:?\/?\/?(?P<domain>[^.]+\.[^\/]+)(?P<path>[^?#]+)?(?P<query>.+)?"


def normalize_url(url: str) -> str | None:
    url = str(url).strip()
    match = re.match(URL_NORMALIZE_REGEX, url)
    if match is None:
        return None
    return f"{match.group('protocol') or 'https'}://{match.group('domain')}{match.group('path') or ''}{match.group('query') or ''}"


def retry(
    max_retry: int,
    retry_interval: int = 0,
    retry_on: Callable[[Exception], bool] = lambda _: True,
) -> Callable[..., Any]:
    def decorator(func):
        def wrapper(*args, **kwargs):
            for i in range(max_retry):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if not retry_on(e):
                        raise e
                    logger.debug(f"Retrying {func.__name__} ({i + 1}/{max_retry})")
                    time.sleep(retry_interval)
                    continue
            raise Exception(
                f"Failed to execute {func.__name__} after {max_retry} retries"
            )

        return wrapper

    return decorator


def retry_async(
    max_retry: int,
    retry_interval: int = 0,
    retry_on: Callable[[Exception], bool] = lambda _: True,
) -> Callable[..., Any]:
    def decorator(func):
        async def wrapper(*args, **kwargs):
            e: Exception | None = None
            for i in range(max_retry):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    if not retry_on(e):
                        raise e
                    logger.debug(f"Retrying {func.__name__} ({i + 1}/{max_retry})")
                    await asyncio.sleep(retry_interval)
                    continue
            raise Exception(
                f"Failed to execute {func.__name__} after {max_retry} retries"
            )

        return wrapper

    return decorator


def cache(func: Callable[..., Any]) -> Callable[..., Any]:
    cache = {}

    def wrapper(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key in cache:
            return cache[key]
        result = func(*args, **kwargs)
        cache[key] = result
        return result

    return wrapper


def cache_async(func: Callable[..., Any]) -> Callable[..., Any]:
    cache = {}

    async def wrapper(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key in cache:
            return cache[key]
        result = await func(*args, **kwargs)
        cache[key] = result
        return result

    return wrapper



import json
import re
from typing import List, TypedDict

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Nicovideo(SiteVisitor):
    NAME = "Nicovideo"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(sp\.)?nicovideo\.jp/(?P<path>user|mylist)/(?P<id>\d+)",
        re.IGNORECASE,
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        if match.group("path") == "mylist":
            res = await session.get(
                f'https://www.nicovideo.jp/mylist/{match.group("id")}'
            )
            return await self.normalize(str(res.url))
        return f'nicovideo.jp/user/{match.group("id")}'

    async def visit(self, url, context: Context, path: str, id: str):
        url = f"https://www.nicovideo.jp/user/{id}"
        res = await session.get(
            f"https://www.nicovideo.jp/user/{id}",
        )
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        element: bs4.Tag = soup.find(attrs={"id": "js-initial-userpage-data"})  # type: ignore
        if element is None:
            logger.warning(f"[Nicovideo] {id} not found")
            return None

        info: Root = json.loads(element.attrs["data-initial-data"])
        user = info["state"]["userDetails"]["userDetails"]["user"]
        context.create_result(
            "Nicovideo",
            url=url,
            score=1.0,
            name=user["nickname"],
            description=user["description"],
            profile_picture=user["icons"]["large"],
        )

        for link in user["sns"]:
            context.visit(link["url"])


class UserLevel(TypedDict):
    currentLevel: int
    nextLevelThresholdExperience: int
    nextLevelExperience: int
    currentLevelExperience: int


class SnsItem0(TypedDict):
    type: str
    label: str
    iconUrl: str
    screenName: str
    url: str


class CoverImage(TypedDict):
    ogpUrl: str
    pcUrl: str
    smartphoneUrl: str


class Icons(TypedDict):
    small: str
    large: str


class User(TypedDict):
    description: str
    decoratedDescriptionHtml: str
    strippedDescription: str
    isPremium: bool
    registeredVersion: str
    followeeCount: int
    followerCount: int
    userLevel: UserLevel
    userChannel: None
    isNicorepoReadable: bool
    sns: List[SnsItem0]
    coverImage: CoverImage
    id: int
    nickname: str
    icons: Icons


class FollowStatus(TypedDict):
    isFollowing: bool


class UserDetails1(TypedDict):
    type: str
    user: User
    followStatus: FollowStatus


class UserDetails(TypedDict):
    userDetails: UserDetails1


class State(TypedDict):
    userDetails: UserDetails


class Root(TypedDict):
    state: State
    nvapi: List



import json
import re
from typing import Dict, List, Optional, TypeAlias, TypedDict

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Pixiv(SiteVisitor):
    NAME = "Pixiv"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"pixiv\.net/users/(?P<id>\d+)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://pixiv.net/users/{match.group("id")}'

    async def visit(self, url, context: Context, id: str):
        res = await session.get(
            f"https://pixiv.net/users/{id}",
        )
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        meta_element: bs4.Tag = soup.find(
            "meta", attrs={"name": "preload-data", "id": "meta-preload-data"}
        )  # type: ignore
        if meta_element is None:
            logger.warning(f"[Pixiv] meta-preload-data not found: {id}")
            return
        info: Root = json.loads(meta_element.attrs["content"])

        if len(info["user"].values()) > 1:
            logger.warning(f"[Pixiv] User is must be unique: {id}")
            return

        for user in info["user"].values():
            context.create_result(
                "Pixiv",
                url=url,
                name=user["name"],
                score=1.0,
                description=user["comment"],
                profile_picture=user["imageBig"],
            )
            if "webpage" in user and user["webpage"] is not None:
                context.visit(user["webpage"])
            if user["social"]:
                for link in user["social"].values():
                    context.visit(link["url"])

            resp = await session.get(
                f"https://sketch.pixiv.net/api/pixiv/user/posts/latest?user_id={id}"
            )
            if resp.status == 200:
                data = resp.json()["data"]
                context.visit(data["user"]["url"])


class Background(TypedDict):
    repeat: None
    color: None
    url: str
    isPrivate: bool


class Social(TypedDict):
    url: str


class Region(TypedDict):
    name: str
    region: str
    prefecture: str
    privacyLevel: str


class Age(TypedDict):
    name: Optional[str]
    privacyLevel: Optional[str]


class UserDetail(TypedDict):
    userId: str
    name: str
    image: str
    imageBig: str
    premium: bool
    isFollowed: bool
    isMypixiv: bool
    isBlocking: bool
    background: Background
    sketchLiveId: None
    partial: int
    acceptRequest: bool
    sketchLives: List
    following: int
    mypixivCount: int
    followedBack: bool
    comment: str
    commentHtml: str
    webpage: str
    social: Dict[str, Social]
    canSendMessage: bool
    region: Region
    age: Age
    birthDay: Age
    gender: Age
    job: Age
    workspace: None
    official: bool
    group: None


User: TypeAlias = Dict[str, UserDetail]


class Root(TypedDict):
    timestamp: str
    user: User



import json
import re
from typing import Tuple
from urllib import parse

import bs4

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor

from .types import thumbnails, ytinitialdata
from .types.about import AboutRes


class Youtube(SiteVisitor):
    NAME = "Youtube"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"((m|gaming)\.)?(youtube\.com|youtu\.be)", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str | None:
        uri = parse.urlparse(url)
        if uri.hostname == "youtu.be":
            return await self._channel_by_video(uri.path[1:])
        type = next(filter(None, uri.path.split("/")))
        if type.startswith("@"):
            return f"https://www.youtube.com/{type}"
        if type == "playlist":
            return None
        if type == "watch":
            return await self._channel_by_video(parse.parse_qs(uri.query)["v"][0])
        if type in ("channel", "user", "c"):
            return await self._channel_by_url(url)
        return url

    async def _channel_by_video(self, video_id: str) -> str | None:
        res = await session.get(
            f"https://www.youtube.com/oembed?url=https://www.youtube.com/watch?v={video_id}&format=json"
        )
        if res.status == 404:
            return None
        data = await res.json()
        return data["author_url"]

    async def _channel_by_url(self, url: str) -> str | None:
        res = await session.get(url)
        if res.status == 404:
            return None
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        element = soup.select_one("#channel-handle")
        if element is None:
            return None
        return f"https://www.youtube.com/{element.text}"

    def parse_thumbnail(self, thumbnails: thumbnails) -> str:
        size = 0
        url: str | None = None
        for thumbnail in thumbnails["thumbnails"]:
            if thumbnail["width"] > size:
                size = thumbnail["width"]
                url = thumbnail["url"]
        if url is None:
            raise RuntimeError("Thumbnail not found")
        return url

    def parse_token(
        self, data: ytinitialdata
    ) -> Tuple[str | None, Tuple[str, str] | None]:
        # TODO: 地獄
        first_link = None
        c4TabbedHeaderRenderer = data["header"]["c4TabbedHeaderRenderer"]
        if "headerLinks" not in c4TabbedHeaderRenderer:
            return (first_link, None)
        channelHeaderLinksViewModel = c4TabbedHeaderRenderer["headerLinks"][
            "channelHeaderLinksViewModel"
        ]
        if "firstLink" in channelHeaderLinksViewModel:
            first_link = channelHeaderLinksViewModel["firstLink"]["commandRuns"][0][
                "onTap"
            ]["innertubeCommand"]["urlEndpoint"]["url"]
        if "more" not in channelHeaderLinksViewModel:
            return (first_link, None)
        runs = channelHeaderLinksViewModel["more"]["commandRuns"]
        command = runs[0]["onTap"]["innertubeCommand"]
        if "showEngagementPanelEndpoint" not in command:
            raise RuntimeError("token not found")
        contents1 = command["showEngagementPanelEndpoint"]["engagementPanel"][
            "engagementPanelSectionListRenderer"
        ]["content"]["sectionListRenderer"]["contents"]
        contents2 = contents1[0]["itemSectionRenderer"]["contents"]
        endpoint = contents2[0]["continuationItemRenderer"]["continuationEndpoint"]
        api_url = endpoint["commandMetadata"]["webCommandMetadata"]["apiUrl"]
        token = endpoint["continuationCommand"]["token"]
        return (first_link, (api_url, token))

    def parse_redirect(self, url: str) -> str:
        uri = parse.urlparse(url)
        if uri.hostname != "www.youtube.com":
            return url
        if uri.path == "/redirect":
            return parse.parse_qs(uri.query)["q"][0]
        return url

    async def visit(self, url: str, context: Context):
        res = await session.get(url)
        if res.status // 100 != 2:
            raise RuntimeError(f"HTTP Error: {res.status}")
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        for script in soup.select("script"):
            if script.string is None:
                continue
            match = re.search(r"ytInitialData\s*=\s*({.*?});", script.string)
            if match is not None:
                data: ytinitialdata = json.loads(match.group(1))
                break
        else:
            raise RuntimeError("ytInitialData not found")
        vanity_id = data["metadata"]["channelMetadataRenderer"]["vanityChannelUrl"]
        name = data["metadata"]["channelMetadataRenderer"]["title"]
        description = data["metadata"]["channelMetadataRenderer"]["description"]
        profile_picture = self.parse_thumbnail(
            data["metadata"]["channelMetadataRenderer"]["avatar"]
        )
        context.create_result(
            site_name="Youtube",
            url=f"https://www.youtube.com/@{vanity_id.split('@')[1]}",
            name=name,
            description=description,
            profile_picture=profile_picture,
        )

        first_link, api = self.parse_token(data)
        if first_link is not None:
            context.visit(self.parse_redirect(first_link))
        if api is None:
            return
        api_url, token = api
        about_res = await session.post(
            f"https://www.youtube.com{api_url}",
            data=json.dumps(
                {
                    "context": {
                        "client": {
                            "userAgent": session.headers["User-Agent"],
                            "clientName": "WEB",
                            "clientVersion": "2.20231219.04.00",
                        }
                    },
                    "continuation": token,
                }
            ),
        )
        about_data: AboutRes = await about_res.json()
        links = about_data["onResponseReceivedEndpoints"][0][
            "appendContinuationItemsAction"
        ]["continuationItems"][0]["aboutChannelRenderer"]["metadata"][
            "aboutChannelViewModel"
        ]["links"]
        for link in links:
            link_url = link["channelExternalLinkViewModel"]["link"]["content"]
            context.visit(self.parse_redirect(link_url))



import asyncio

import iwashi


async def main():
    result = await iwashi.visit("https://www.youtube.com/@hololive")
    if result:
        iwashi.helper.print_result(result)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
    loop.close()



import re

import bs4

from iwashi.helper import HTTP_REGEX, session
from iwashi.visitor import Context, SiteVisitor


class Booth(SiteVisitor):
    NAME = "Booth"
    URL_REGEX: re.Pattern = re.compile(
        HTTP_REGEX + r"(?P<id>[\w-]+)\.booth\.pm", re.IGNORECASE
    )

    async def normalize(self, url: str) -> str:
        match = self.URL_REGEX.match(url)
        if match is None:
            return url
        return f'https://{match.group("id")}.booth.pm'

    async def visit(self, url, context: Context, id: str):
        res = await session.get(f"https://{id}.booth.pm")
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        name_element: bs4.Tag = soup.find(
            attrs={"class": "home-link-container__nickname"}
        )  # type: ignore
        name = name_element is not None and name_element.find("a") or None
        desc_element: bs4.Tag = soup.find(attrs={"class": "description"})  # type: ignore
        description = (
            desc_element is not None
            and desc_element.find(attrs={"v-html": "formattedDescription"})
            or None
        )
        avater_element: bs4.Tag = soup.find(attrs={"class": "avatar-image"})  # type: ignore

        context.create_result(
            "Booth",
            url=url,
            name=name.text if name is not None else None,
            score=1.0,
            description=description.text if description is not None else None,
            profile_picture=avater_element.attrs["style"].split("url(")[1].split(")")[0]
            if avater_element is not None
            else None,
        )

        for link in desc_element.find_all(attrs={"class": "shop-contacts__link"}):
            link: bs4.Tag = link.find("a")  # type: ignore
            if link.has_attr("href"):
                context.visit(link.attrs["href"])
